unit QT;

(*
  QT warper for Kylix (c)2001 by Paul TOTH <tothpaul@free.fr>
  http://tothpaul.free.fr
*)

interface

uses
 libqt,SysUtils;

Const // QFont::Weight
 Light    = 25;
 Normal   = 50;
 DemiBold = 63;
 Bold     = 75;
 Black    = 87;

Type
 TVTable=array[word] of pointer;
 PVTable=^TVTable;

 TVTHandler=record
  count :integer;
  vtable:TVTable;
 end;
 PVTHandler=^TVTHandler;

 TWrapper=record
  Kylix:TObject;
  Linux:byte;
 end;


Type
 QWidget=class;

 QBase=class
 protected
  wrap:^TWrapper;
  this:QHandle;   // QT object instance
  vtable:pointer; // QObject vtable
  handler:PVTHandler;
  function InstallHandler(var VTHandler:PVTHandler):boolean; virtual;
 public
  Constructor CreateSize(Size:integer);
  Destructor Destroy; override;
  property Handle:QHandle read this;
 end;

 QFont=class(QBase)
 public
  Constructor Create(family:pchar; pointSize :integer = 12; weight:integer=Normal; italic:boolean=false);
 end;

 QObject=class(QBase)
 public
  class procedure connect(sender:QObject; signal:string; receiver:QObject; member:string);
  procedure SetName(Name:pchar);
 end;

 QApplication=class(QObject)
 public
  Constructor Create;
  procedure SetMainWidget(Widget:QWidget);
  procedure exec;
 end;
{
 QPaintDevice=class
 end;
}
 QWidget=class(QObject)
 protected
  function InstallHandler(var VTHandler:PVTHandler):boolean; override;
 public
  Constructor Create(parent:QWidget=nil; name:pchar=nil; flags:integer=0);
 // wrapper
  procedure PaintEvent(QPaintEvent:pointer); virtual;
  procedure ResizeEvent(QResizeEvent:pointer); virtual;
 // static
  procedure SetMinimumSize(width,height:integer);
  procedure SetMaximumSize(width,height:integer);
  function x:integer;
  function y:integer;
  function width:integer;
  function height:integer;
 // virtual
  procedure SetFont(Font:QFont);
  procedure SetStyle(Style:TGUIStyle);
  procedure Show;
  procedure Move(left,top:integer);
  procedure Resize(width,height:integer);
  procedure SetGeometry(Left,Top,Width,Height:integer);
 end;

 QButton=class(QWidget)
 end;

 QPushButton=class(QButton)
 public
  Constructor Create(Text:string; Parent:QWidget = nil; Name:pchar=nil);
 end;

Const
 QScrollBar_Horizontal = 0;
 QScrollBar_Vertital   = 1;
Type
 QScrollBar=class(QWidget) // and QRangeControl
 public
  Constructor Create(min,max,linestep,pagestep,value,orientation:integer; Parent:QWidget = nil; Name:pchar=nil);
 end;

 QFrame=class(QWidget)
 end;

 QLCDNumber=class(QFrame)
 public
  Constructor Create(numDigits:integer; Parent:QWidget = nil; Name:pchar=nil);
 end;

var
 qApp:QApplication;

implementation

// vtable
var
 QApplicationVT:PVTHandler=nil;
 QWidgetVT     :PVTHandler=nil;
 QPushButtonVT :PVTHandler=nil;
 QScrollBarVT  :PVTHandler=nil;
 QLCDNumberVT  :PVTHandler=nil;

function VTSize(vt:pointer):integer;
 var
  vtable:PVTable absolute vt;
 begin
  result:=1; // first entry (0) is null
  while vtable[result]<>nil do inc(result); // I think vtable end with a null pointer...
  result:=(result+1)*sizeof(pointer);
 end;

procedure ReleaseVT(var VT:PVTHandler);
 begin
  dec(VT.Count);
  if VT.Count=0 then begin
   Freemem(VT);
   VT:=nil;
  end;
 end;

// VT Handler

procedure DestroyHandler(this:pointer); cdecl;
 var
  W:^TWrapper;
 begin
  W:=pointer(integer(this)-4);
  write(' DestroyHandler -> ');
  QObject(W.Kylix).Free;
 end;

procedure PaintHandler(this,Event:pointer); cdecl;
 var
  W:^TWrapper;
 begin
  W:=pointer(integer(this)-4);
  writeln(' PaintHandler -> ',QObject(W.Kylix).ClassName,'.PaintEvent');
  QWidget(W.Kylix).PaintEvent(Event);
 end;

procedure ResizeHandler(this,Event:pointer); cdecl;
 var
  W:^TWrapper;
 begin
  W:=pointer(integer(this)-4);
  writeln(' ResizeHandler -> ',QObject(W.Kylix).ClassName,'.ResizeEvent');
  QWidget(W.Kylix).ResizeEvent(Event);
 end;

Constructor QBase.CreateSize(Size:integer);
 begin
  inherited Create;
  GetMem(wrap,Size+sizeof(self));
  wrap.Kylix:=self;
  this:=@wrap.linux;
 end;

Destructor QBase.Destroy;
 begin
  write(ClassName,'.Destroy');
  if Assigned(PQObjectVT(vtable).Destroy) then begin
   writeln(' -> QDestroy ',integer(@PQObjectVT(vtable).Destroy));
   PQObject(this).vtable:=vtable;
   try
    PQObjectVT(vtable).Destroy(this);
   except
    writeln(' erreur de destruction...');
   end;
  end else begin
   writeln;
  end;
  FreeMem(wrap); // and "this"
  ReleaseVT(Handler);
  inherited;
 end;

function QBase.InstallHandler(var VTHandler:PVTHandler):boolean;
 var
  size:integer;
 begin
  vtable:=PQObject(this).vtable; // save old vtable
  result:=(VTHandler=nil);
  if result then begin
   size:=vtsize(vtable);
   getmem(VTHandler,size+4);
   VTHandler.count:=0;
   move(vtable^,VTHandler.vtable,size);
   PQObjectVT(@VTHandler.vtable).Destroy:=DestroyHandler; 
  end;
  handler:=VTHandler;
  inc(VTHandler.Count);
  PQObject(this).vtable:=@VTHandler.vtable;
 end;

Constructor QFont.Create(family:pchar; pointSize :integer = 12; weight:integer=Normal; italic:boolean=false);
 begin
  inherited CreateSize(8);
  Create_QFont(this,family,pointSize,weight,italic);
 end;

class procedure QObject.connect(sender:QObject; signal:string; receiver:QObject; member:string);
 begin
  QObject_Connect(sender.this,pchar('2'+signal),receiver.this,pchar('1'+member));
 end;

procedure QObject.SetName(Name:PChar);
 begin
  QObject_SetName(this,Name);
 end;

Constructor QApplication.Create;
 begin
  inherited CreateSize(52);
  Create_QApplication(this,@ArgCount,ArgValues);
  InstallHandler(QApplicationVT);
  qApp:=Self;
 end;

procedure QApplication.SetMainWidget(Widget:QWidget);
 begin
  QApplication_setMainWidget(this,Widget.this);
 end;

procedure QApplication.exec;
 begin
  QApplication_exec(this);
 end;

Constructor QWidget.Create(Parent:QWidget=nil; name:pchar=nil; flags:integer=0);
 begin
  inherited CreateSize(104);
  if parent=nil then
   Create_QWidget(this,nil,name,flags)
  else begin
   Create_QWidget(this,parent.this,name,flags);
  end;
  InstallHandler(QWidgetVT);
 end;

Function QWidget.InstallHandler(Var VTHandler:PVTHandler):boolean;
 begin
  result:=inherited InstallHandler(VTHandler);
  if Result then begin
   PQWidgetVT(@VTHandler.vtable).PaintEvent :=PaintHandler;
   PQWidgetVT(@VTHAndler.vtable).ResizeEvent:=ResizeHandler;
  end;
 end;

procedure QWidget.PaintEvent(QPaintEvent:pointer);
 begin
  if Assigned(PQWidgetVT(vtable).PaintEvent) then PQWidgetVT(vtable).PaintEvent(this,QPaintEvent);
 end;

procedure QWidget.ResizeEvent(QResizeEvent:pointer);
 begin
  if Assigned(PQWidgetVT(vtable).ResizeEvent) then PQWidgetVT(vtable).ResizeEvent(this,QResizeEvent);
 end;

Procedure QWidget.SetMinimumSize(width,height:integer);
 begin
  QWidget_setMinimumSize(this,width,height);
 end;

Procedure QWidget.SetMaximumSize(width,height:integer);
 begin
  QWidget_setMaximumSize(this,width,height);
 end;

function QWidget.x:integer;
 begin
  result:=PQWidget(this).frect.x1;
 end;

function QWidget.y:integer;
 begin
  result:=PQWidget(this).frect.y1;
 end;

function QWidget.width:integer;
 begin
  with PQWidget(this).crect do result:=x2-x1;
 end;

function QWidget.height:integer;
 begin
  with PQWidget(this).crect do result:=y2-y1;
 end;

Procedure QWidget.SetFont(Font:QFont);
 begin
  if assigned(PQWidgetVT(vtable).SetFont) then PQWidgetVT(vtable).SetFont(this,Font.Handle);
 end;

procedure QWidget.SetStyle(Style:TGUIStyle);
 begin
  if assigned(PQWidgetVT(vtable).SetStyle) then PQWidgetVT(vtable).SetStyle(this,Style);
 end;

Procedure QWidget.Show;
 begin
  PQWidgetVT(vtable).Show(this);
 end;

Procedure QWidget.Resize(width,height:integer);
 begin
  PQWidgetVT(vtable).Resize(this,width,height);
 end;

Procedure QWidget.Move(left,top:integer);
 begin
  PQWidgetVT(vtable).Move(this,left,top);
 end;

procedure QWidget.SetGeometry(Left,Top,Width,Height:integer);
 begin
  PQWidgetVT(vtable).SetGeometry(this,left,top,width,height);
 end;

Constructor QPushButton.Create(text:string; parent:QWidget = nil; name:pchar=nil);
 begin
  inherited CreateSize(128);
  if Parent=nil then
   Create_QPushButton(this,pchar(text),nil,name)
  else begin
   Create_QPushButton(this,pchar(text),parent.this,name)
  end;
  InstallHandler(QPushButtonVT);
 end;

Constructor QScrollBar.Create(min,max,linestep,pagestep,value,orientation:integer; Parent:QWidget = nil; Name:pchar=nil);
 begin
  inherited CreateSize(144);
  if Parent=nil then
   Create_QScrollBar(this,min,max,linestep,pagestep,value,orientation,nil,name)
  else begin
   Create_QScrollBar(this,min,max,linestep,pagestep,value,orientation,parent.this,name)
  end;
  InstallHandler(QScrollBarVT);
 end;

Constructor QLCDNumber.Create(numDigits:integer; Parent:QWidget = nil; Name:pchar=nil);
 begin
  inherited CreateSize(156);
  if Parent=nil then
   Create_QLCDNumber(this,numDigits,nil,name)
  else begin
   Create_QLCDNumber(this,numDigits,parent.this,name)
  end;
  InstallHandler(QLCDNumberVT);
 end;

initialization
 writeln('Debut de l''application');
 qApp:=nil;
// InitVMT;

finalization
 writeln('Destruction de l''application en cours..');
 writeln(' Destroy_QApplication=',integer(@Destroy_QApplication));
 writeln(' Destroy_QWidget=',integer(@Destroy_QWidget));
 writeln(' Destroy_QPushButton=',integer(@Destroy_QPushButton));
 //qApp.free;
 PQObjectVT(@QApplicationVT.vtable).Destroy(qApp.Handle);
 writeln('bye.');

end.


